﻿using Components.Aphid.Interpreter;
using Components.Aphid.Lexer;
using Components.Aphid.Library;
using Components.Aphid.Parser;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;

namespace AphidUI.ViewModels
{
    public class WindowViewModel : ViewModel
    {
        public const string NoValue = "Expression has been evaluated and has no value";

        private CodeViewer _codeViewer;

        private AphidInterpreter _interpreter = new AphidInterpreter();

        public AphidInterpreter Interpreter
        {
            get { return _interpreter; }
        }

        private List<string> _codeHistory = new List<string>();

        private int _codeHistoryIndex = 0;

        private int _maxQueueSize = 32;

        private string _code;

        public string Code
        {
            get { return _code; }
            set { SetProperty(ref _code, value); }
        }

        private string _output;

        public string Output
        {
            get { return _output; }
            set { SetProperty(ref _output, value); }
        }

        private ObservableCollection<VariableViewModel> _variables = new ObservableCollection<VariableViewModel>();

        public ObservableCollection<VariableViewModel> Variables
        {
            get { return _variables; }
        }

        private ObservableCollection<ExpressionViewModel> _expressions = new ObservableCollection<ExpressionViewModel>();

        public ObservableCollection<ExpressionViewModel> Expressions
        {
            get { return _expressions; }
        }

        public WindowViewModel(CodeViewer outputTextBox)
        {
            _codeViewer = outputTextBox;
        }

        public void Init()
        {
            _expressions.CollectionChanged += Expressions_CollectionChanged;
            _codeViewer.Document.Blocks.Clear();
        }

        private void Expressions_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (e.NewItems != null)
            {
                foreach (ExpressionViewModel i in e.NewItems)
                {
                    i.ExpressionChanged += Item_ExpressionChanged;
                }
            }

            if (e.OldItems != null)
            {
                foreach (ExpressionViewModel i in e.OldItems)
                {
                    i.ExpressionChanged -= Item_ExpressionChanged;
                }
            }
        }

        private void Item_ExpressionChanged(object sender, EventArgs e)
        {
            ExecuteExpression((ExpressionViewModel)sender);
        }

        private void AddHistoricalCode()
        {
            _codeHistory.Add(Code);

            if (_codeHistory.Count > _maxQueueSize)
            {
                _codeHistory.RemoveAt(0);
            }

            _codeHistoryIndex = _codeHistory.Count;
        }

        private string GetHistoricalCode(Func<bool> changeAndCheck, Func<int> resetIndex)
        {
            if (!_codeHistory.Any())
            {
                return null;
            }
            else if (changeAndCheck())
            {
                _codeHistoryIndex = resetIndex();
            }

            return _codeHistory[_codeHistoryIndex];
        }

        private string GetPreviousCode()
        {
            return GetHistoricalCode(
                () => (--_codeHistoryIndex) == -1,
                () => _codeHistory.Count - 1);
        }

        private string GetNextCode()
        {
            return GetHistoricalCode(
                () => (++_codeHistoryIndex) >= _codeHistory.Count,
                () => _codeHistoryIndex = 0);
        }

        private void UpdateVariables()
        {
            Variables.Clear();

            foreach (var v in _interpreter.CurrentScope.Select(x => new VariableViewModel()
            {
                Name = x.Key,
                Value = x.Value.Value,
            }))
            {
                Variables.Add(v);
            }
        }

        private void ExecuteCode()
        {
            AddHistoricalCode();

            var tokens = new AphidLexer(Code + " ").GetTokens();

            if (!tokens.Any(x => x.TokenType == AphidTokenType.EndOfStatement))
            {
                tokens.Add(new AphidToken(AphidTokenType.EndOfStatement, "", 0));
            }

            List<Components.Aphid.Parser.Expression> ast = null;

            try
            {
                ast = new AphidParser(tokens).Parse();
            }
            catch (AphidParserException ex)
            {
                _codeViewer.AppendParserException(Code, ex);

                return;
            }

            if (ast.Count != 1)
            {
                throw new InvalidOperationException();
            }

            var exp = ast.First();
            var unary = exp as UnaryOperatorExpression;

            if (unary == null ||
                unary.Operator != AphidTokenType.retKeyword)
            {
                ast.Clear();
                ast.Add(new UnaryOperatorExpression(
                    AphidTokenType.retKeyword,
                    exp));
            }

            var serializer = new AphidSerializer() { IgnoreFunctions = false };

            try
            {
                _interpreter.Interpret(ast);
            }
            catch (AphidRuntimeException ex)
            {
                _codeViewer.AppendException(Code, "Runtime error", ex);

                return;
            }
            catch (AphidParserException ex)
            {
                _codeViewer.AppendParserException(Code, ex);

                return;
            }

            var retVal = _interpreter.GetReturnValue();
            _codeViewer.AppendOutput(Code, serializer.Serialize(retVal));

            UpdateVariables();
            ExecuteExpressions();
        }

        private void ExecuteExpression(ExpressionViewModel vm)
        {
            try
            {
                _interpreter.Interpret("ret " + vm.Expression + ";");
            }
            catch (AphidRuntimeException e)
            {
                vm.Value = "Runtime exception: " + e.Message;
                return;
            }
            catch (AphidParserException e)
            {
                vm.Value = "Parser exception: " + e.Message;
                return;
            }

            var retVal = new AphidSerializer() { IgnoreFunctions = false }
                .Serialize(_interpreter.GetReturnValue());
            vm.Value = retVal;
        }

        private void ExecuteExpressions()
        {
            foreach (var exp in _expressions)
            {
                ExecuteExpression(exp);
            }
        }

        public void MainWindow_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.Key)
            {
                case Key.F5:
                    ExecuteCode();
                    break;
            }
        }

        public void InputTextBox_KeyDown(object sender, KeyEventArgs e)
        {
            var tb = (TextBox)sender;

            switch (e.Key)
            {
                case Key.Enter:
                    ExecuteCode();
                    Code = "";
                    break;

                case Key.Up:
                    Code = GetPreviousCode();
                    tb.ScrollToEnd();
                    break;

                case Key.Down:
                    Code = GetNextCode();
                    tb.ScrollToEnd();
                    break;
            }
        }

        public void ClearCodeHandler(object sender, RoutedEventArgs e)
        {
            Code = "";
            _codeViewer.Document.Blocks.Clear();
        }

        public void ResetInterpreterHandler(object sender, RoutedEventArgs e)
        {
            Code = "";
            _codeViewer.Document.Blocks.Clear();
            _interpreter = new AphidInterpreter();
            _variables.Clear();
        }
    }
}
